#!/usr/bin/python
# vim: encoding=utf-8

# @(#) Quake Server Poll
# @(#) iekko <iekko@free.fr>
# @(#) $Id: qsp.py,v 1.22 2009/01/28 19:55:42 iekko Exp $

from twisted.internet import reactor, defer
from twisted.internet.protocol import DatagramProtocol
from getopt import getopt, GetoptError
from os import mkdir, path
from re import compile, sub
import socket
from string import maketrans
from sys import argv, exit, stderr, stdout
from time import gmtime, strftime, time
from xml.dom.minidom import Document, parseString

# Informations
BASENAME = path.basename(argv[0])
NAME = "QSP"
DESC = "Quake Servers Poll"
VERSION = "0.3"
AUTHOR = "iekko <iekko@free.fr>"
WEBPAGE = "http://iekko.free.fr/devel/qsp/"
# Configuration
HOMEDIR = path.join(path.expanduser("~"), ".qsp")
CONF = {}
PROPERTIES = ("cache_favorites", "cache_server", "cache_servers",
		"filter", "master", "protocol", "port_master", "port_server",
		"timeout", "format", "sort")
# Formats
ALL = maketrans("", "")
NON_PRINTABLE = ALL[:32] + ALL[0] * (127 - 32) + ALL[127:256]
POSITIVE_INTEGER = compile("^\d+$")
NON_NEGATIVE_INTEGER = compile("^[1-9]\d*$")
NON_NEGATIVE_FLOAT = compile("^([1-9]\d*|0*\.\d*[1-9]\d*|0*\d*[1-9]\d*\.\d*)$")
FORMAT = compile("^(text|color|raw|xml)$")
IGNORE = compile("^\s*(#|$)")
PROPERTY = compile("^\s*([^\s]+)\s*=\s*([^\s]+)\s*$")
ADDRESS = compile("^([^:]+)(:(\d+))?$")
SORT = compile("^\+?(server|address|port|map|players|slots|ping)$")

class QError(Exception):
	"""Common error"""

	def __init__(self, message):
		self.message = message

	def __str__(self):
		return self.message

class Host:
	"""Host"""

	def __init__(self, address, port):
		self.address = address
		self.port = port

	def __str__(self):
		return "(%s, %d)" % (self.address, self.port)

	def new_from_key(key):
		return Host(key[0], key[1])
	new_from_key = staticmethod(new_from_key)

	def get_key(self):
		return (self.address, self.port)

class Server(Host):
	"""Server"""

	NAME = "sv_hostname"
	GAME = "gamename"
	MAP = "mapname"
	VERSION = "version"
	PROTOCOL = "protocol"
	TIMELIMIT = "timelimit"
	MAX = "sv_maxclients"
	PRIVATE = "sv_privateClients"
	PING = ".ping"
	PLAYERS = ".players"
	SLOTS = ".slots"
	IGNORE = (PING, PLAYERS, SLOTS)
	PING999 = "999"

	def __init__(self, address, port, properties, players):
		Host.__init__(self, address, port)
		self.properties = properties
		self.players = players

	def __str__(self):
		return "(%s, %d, %s, %s)" % (self.address, self.port,
				self.properties, map(str, self.players))

	def new_from_key(key, properties, players):
		return Server(key[0], key[1], properties, players)
	new_from_key = staticmethod(new_from_key)

	def is_empty(self):
		"""Check if the server is empty"""
		return len(self.players) == 0

	def is_full(self):
		"""Check if the server is full"""
		full = False
		if self.properties.has_key(Server.SLOTS):
			if self.properties.has_key(Server.PLAYERS):
				full = self.properties[Server.PLAYERS] \
						>= self.properties[Server.SLOTS]
			else:
				full = len(self.players) >= self.properties[Server.SLOTS]
		return full

class Player:
	"""Player"""

	def __init__(self, num, score, ping, name):
		self.num = num
		self.score = score
		self.ping = ping
		self.name = name

	def __str__(self):
		return "(%d, %d, %s, %s)" % (self.num,
				self.score, self.ping, self.name)

class Protocol(DatagramProtocol):
	"""Protocol"""

	def __init__(self, deferred):
		self.__deferred = deferred
		self.sendings = {}
		self.servers = {}

	def datagramReceived(self, datagram, (address, port)):
		"""Work on a received datagram"""
		now = time()
		if datagram.find("\xFF\xFF\xFF\xFFgetserversResponse") == 0:
			# request each server parsed from master response
			for host in datagram.split("\\")[1:-1]:
				if len(host) == 6:
					self.request_server(Host("%d.%d.%d.%d" % (
						ord(host[0]),
						ord(host[1]),
						ord(host[2]),
						ord(host[3])),
						ord(host[4]) * 256 + ord(host[5])))
		elif datagram.find("\xFF\xFF\xFF\xFFstatusResponse") == 0:
			# cache informations parsed from server response
			properties = {}
			players = []
			server = Server(address, port, properties, players)
			key = server.get_key()
			# check if request has been sent
			if self.sendings.has_key(key):
				sending = self.sendings[key]
				del self.sendings[key]
				records = datagram.split("\x0A")[1:]
				if len(records) > 1:
					# parse properties
					props = records[0].split("\\")[1:]
					for (i, key) in enumerate(props[:-1:2]):
						properties[key] = normalize(props[2 * i + 1])
					# parse players
					for (i, player) in enumerate(records[1:-1]):
						player = player.split(" ")
						try:
							players.append(Player(i,
								int(player[0]), int(player[1]),
								normalize(" ".join(player[2:]))[1:-1]))
						except ValueError:
							pass
					# add properties
					ping = int(round(1000 * (now - sending)))
					properties[Server.PING] = ping < Server.PING999 \
							and ping or Server.PING999
					properties[Server.PLAYERS] = len(players)
					if properties.has_key(Server.MAX):
						try:
							if properties.has_key(Server.PRIVATE):
								properties[Server.SLOTS] = \
										int(properties[Server.MAX]) \
										- int(properties[Server.PRIVATE])
							else:
								properties[Server.SLOTS] = \
										int(properties[Server.MAX])
						except ValueError:
							properties[Server.SLOTS] = 0
					else:
						properties[Server.SLOTS] = 0
					# cache server
					self.servers[server.get_key()] = server
				# end
				if len(self.sendings) == 0:
					self.end()

	def request_servers(self, master, protocol, empty, full):
		"""Request a master for servers
		master		: the master to request
		protocol	: the protocol of servers
		empty		: whether include or not empty servers
		full		: whether include or not full servers"""
		# prepare query
		query = ["\xFF\xFF\xFF\xFFgetservers", str(protocol)]
		if empty:
			query.append("empty")
		if full:
			query.append("full")
		# send query to master
		self.transport.write(" ".join(query), (master.address, master.port))

	def request_server(self, server):
		"""Request a server for properties and players
		server	: the server to request"""
		try:
			# send query to server
			self.transport.write("\xFF\xFF\xFF\xFFgetstatus",
					(server.address, server.port))
			# record sending time
			self.sendings[server.get_key()] = time()
		except Exception, e:
			warn("(%s, %d): error while sending request to server" \
					% (server.address, server.port))

	def end(self):
		"""End"""
		try:
			self.__deferred.callback(self)
		except defer.AlreadyCalledError, e:
			pass

def version():
	"""Show version"""
	stdout.write("%s %s\n%s\nWritten by %s\n%s\n" \
			% (NAME, VERSION, DESC, AUTHOR, WEBPAGE))

def help():
	"""Show help"""
	version()
	stdout.write("Usage: %s [OPTIONS]...\n"
			"  -A, --add-favorite       add server to favorites\n"
			"  -C, --cache              specify cache to use\n"
			"  -c, --color              use text format with ansi color\n"
			"  -D, --delete-favorite    delete server from favorites\n"
			"  -E, --empty              include empty servers\n"
			"  -F, --full               include full servers\n"
			"  -f, --favorites          poll favorites\n"
			"  -g, --game               load game configuration\n"
			"  -h, --help               show help, then exit\n"
			"  -i, --filter             filter with player name regexp\n"
			"  -l, --list               poll servers from master\n"
			"  -m, --master             specify master host\n"
			"  -P, --protocol           specify server protocol\n"
			"  -p, --port               specify port to use\n"
			"  -R, --raw                use raw format\n"
			"  -r, --refresh            poll servers from cache\n"
			"  -S, --sort               sort by a field\n"
			"  -s, --server             poll server\n"
			"  -T, --text               use text format\n"
			"  -t, --properties         display properties\n"
			"  -v, --version            show version, then exit\n"
			"  -W, --timeout            wait time for response\n"
			"  -X, --xml                use XML format\n"
			"  -y, --players            display players\n"
			% BASENAME)

def error(mesg, errno):
	"""Show error and exit
	mesg	: the message to display
	errno	: the error code with which to exit"""
	warn(mesg)
	exit(errno)

def warn(mesg):
	"""Show warning message
	mesg	: the message to show"""
	stderr.write("%s: %s\n" % (BASENAME, mesg))

def info(mesg):
	"""Show information message
	mesg	: the message to show"""
	stdout.write(mesg)

def normalize(line):
	"""Normalize a line
	line	: the line to normalize"""
	return sub("(^\s*|\s*$)", "", line.translate(ALL, NON_PRINTABLE))

def decolorise(line, max=None):
	"""Decolorise a line
	line	: the line to decolorise
	max		: the maximal length"""
	return sub("\^.", "", line)[:max and max or len(line)]

def colorise(line, max):
	"""Colorise a line
	line	: the line to colorise
	max		: the maximal length"""
	result = []
	line = "^7%s^7" % line
	left = max
	for token in line.split("^")[1:]:
		# append ^
		if len(token) == 0:
			if left >= 1:
				result.append("^")
				left -= 1
			else:
				result.append("\x1B[0m")
				max += 4
				break
		else:
			# append color
			color = ord(token[0]) % 7 + 1
			if color == 5:
				color = 6
			elif color == 6:
				color = 5
			if color == 7:
				result.append("\x1B[0m")
				max += 4
			else:
				result.append("\x1B[3%dm" % color)
				max += 5
			# append token
			token = token[1:]
			length = len(token)
			if left >= length:
				result.append(token)
				left -= length
			else:
				result.append(token[:left])
				result.append("\x1B[0m")
				max += 4
				break
	return "".join(result).ljust(max)[:max]

def servers2xml(servers):
	"""Convert servers to xml
	servers	: the servers to convert"""
	doc = Document()
	# create headers
	doc.appendChild(doc.createProcessingInstruction("xml-stylesheet",
		'type="text/xsl" href="html.xsl"'))
	servs = doc.createElement("servers")
	doc.appendChild(servs)
	servs.setAttribute("xmlns",
			"http://iekko.free.fr/devel/qsp")
	servs.setAttribute("xmlns:xsi",
			"http://www.w3.org/2001/XMLSchema-instance")
	servs.setAttribute("xsi:schemaLocation",
			"http://iekko.free.fr/devel/qsp qsp.xsd")
	servs.setAttribute("date", strftime("%Y-%m-%d %H:%M:%S +0000", gmtime()));
	for server in servers:
		# add server
		serv = doc.createElement("server")
		servs.appendChild(serv)
		serv.setAttribute("address", server.address)
		serv.setAttribute("port", str(server.port))
		elems = doc.createElement("properties")
		serv.appendChild(elems)
		# add properties
		for name, value in server.properties.items():
			prop = doc.createElement("property")
			elems.appendChild(prop)
			prop.setAttribute("name", str(name))
			prop.setAttribute("value", str(value))
		elems = doc.createElement("players")
		serv.appendChild(elems)
		# add players
		for player in server.players:
			pl = doc.createElement("player")
			elems.appendChild(pl)
			pl.setAttribute("num", str(player.num))
			pl.setAttribute("score", str(player.score))
			pl.setAttribute("ping", str(player.ping))
			pl.setAttribute("name", player.name)
	return doc.toprettyxml(encoding="utf-8")

def xml2servers(xml):
	"""Convert xml to servers
	xml	: the xml to convert"""
	servers = []
	for serv in parseString(xml).documentElement.getElementsByTagName("server"):
		properties = {}
		players = []
		servers.append(Server(serv.getAttribute("address"),
			int(serv.getAttribute("port")), properties, players))
		# parse properties
		for pr in serv.getElementsByTagName(
				"properties")[0].getElementsByTagName("property"):
			properties[pr.getAttribute("name")] = pr.getAttribute("value")
		# parse players
		for pl in serv.getElementsByTagName(
				"players")[0].getElementsByTagName("player"):
			players.append(Player(
				int(pl.getAttribute("num")),
				int(pl.getAttribute("score")),
				int(pl.getAttribute("ping")),
				pl.getAttribute("name")))
	return servers

def servers2text(servers, properties, players, format):
	"""Convert servers to text
	servers		: the servers to convert
	properties	: whether display or not properties
	players		: whether display or not players
	format		: the output format (text, color, raw or xml)"""
	if format == "xml":
		return servers2xml(servers)
	text = []
	srv_section = "%-30s %-15s %6s %-12s %-7s %4s"
	pr_section = "    %-30s %-44s"
	pr_ignore = [Server.NAME, Server.GAME, Server.MAP, Server.VERSION,
			Server.PROTOCOL, Server.TIMELIMIT,
			Server.PLAYERS, Server.SLOTS, Server.PING]
	pl_section = "    %3s %5s %4s %-60s"
	empty = ""
	if format == "text":
		filter = decolorise
	elif format == "color":
		filter = colorise
	else:
		filter = lambda x, y: x[:y]
	text.append(srv_section % ("SERVER NAME", "ADDRESS", "PORT ",
		"MAP NAME", "PLAYERS", "PING"))
	for server in servers:
		# add server
		p_address = "%s:%d" % (server.address, server.port)
		props = server.properties
		text.append(srv_section % (
			props.has_key(Server.NAME) \
					and filter(props[Server.NAME], 30) or p_address,
			server.address,
			server.port,
			props.has_key(Server.MAP) and props[Server.MAP][:12] or empty,
			"%2s %4s" % (props[Server.PLAYERS], "(%s)" % props[Server.SLOTS]),
			props[Server.PING]))
		# add properties
		if properties:
			text.append(pr_section % ("PROPERTY", "VALUE"))
			text.append(pr_section % ("Name",
				props.has_key(Server.NAME) \
						and filter(props[Server.NAME], 60) or p_address))
			text.append(pr_section % ("Game name",
				props.has_key(Server.GAME) \
						and props[Server.GAME] or empty))
			text.append(pr_section % ("Map",
				props.has_key(Server.MAP) \
						and props[Server.MAP] or empty))
			text.append(pr_section % ("Version",
				props.has_key(Server.VERSION) \
						and props[Server.VERSION] or empty))
			text.append(pr_section % ("Protocol",
				props.has_key(Server.PROTOCOL) \
						and props[Server.PROTOCOL] or empty))
			text.append(pr_section % ("Time limit",
				props.has_key(Server.TIMELIMIT) \
						and props[Server.TIMELIMIT] or empty))
			keys = props.keys()
			keys.sort()
			for key in keys:
				if key not in pr_ignore:
					text.append(pr_section % (key, props[key]))
		# add players
		if players:
			if not server.is_empty():
				text.append(pl_section % ("NUM", "SCORE", "PING", "NAME"))
				for player in server.players:
					text.append(pl_section % (player.num, player.score,
						player.ping, filter(player.name, 60)))
	text.append(empty)
	return "\n".join(text)

def sort_servers(servers, field=None):
	"""Sort servers using a field
	servers	: the servers to sort
	field	: the field to use"""
	if field:
		if field[0] == "+":
			field = field[1:]
			order = -1
		else:
			order = 1
		if field == "server":
			servers.sort(lambda x,y: order * cmp(
				x.properties.has_key(Server.NAME) \
						and x.properties[Server.NAME] \
						or "%s:%d" % (x.address, x.port),
				y.properties.has_key(Server.NAME) \
						and y.properties[Server.NAME] \
						or "%s:%d" % (y.address, y.port)))
		elif field == "ping":
			servers.sort(lambda x,y: order * cmp(
				x.properties[Server.PING], y.properties[Server.PING]))
		elif field == "players":
			servers.sort(lambda x,y: order * cmp(
				x.properties[Server.PLAYERS], y.properties[Server.PLAYERS]))
		elif field == "slots":
			servers.sort(lambda x,y: order * cmp(
				x.properties[Server.SLOTS], y.properties[Server.SLOTS]))
		elif field == "map":
			servers.sort(lambda x,y: order * cmp(
				x.properties[Server.MAP], y.properties[Server.MAP]))
		elif field == "address":
			servers.sort(lambda x,y: order * cmp(x.address, y.address))
		elif field == "port":
			servers.sort(lambda x,y: order * cmp(x.port, y.port))

def store_servers(file, servers):
	"""Store servers to file
	file	: the file to which store the servers
	servers	: the servers to store"""
	xml = servers2xml(servers)
	try:
		f = open(file, "w")
	except IOError, e:
		raise QError(str(e))
	try:
		f.write(xml)
	except IOError, e:
		f.close()
		raise QError(str(e))
	f.close()

def load_servers(file):
	"""Load servers from file
	file	: the file from which to read servers"""
	try:
		f = open(file)
	except IOError, e:
		raise QError(str(e))
	try:
		xml = f.read()
	except IOError, e:
		f.close()
		raise QError(str(e))
	f.close()
	return xml2servers(xml)

def digest_servers(protocol, cache, purge, empty, full,
		filter, format, properties, players, sort):
	"""Digest the informations received from servers
	protocol	: the protocol from which to digest servers
	cache		: the cache on which to store servers
	purge		: whether purge or not unreached servers
	empty		: whether display or not empty servers
	full		: whether display or not full servers
	filter		: the player name regexp to apply as filter on servers
	format		: the chosen output format
	properties	: whether display or not properties
	players		: whether display or not players
	sort		: the sort to use"""
	protocol.transport.stopListening()
	reactor.stop()
	servers = protocol.servers.values()
	# cache servers
	if not purge:
		for key in protocol.sendings.keys():
			servers.append(Server.new_from_key(key, {Server.PING: 999,
				Server.PLAYERS: 0, Server.SLOTS: 0, }, []))
	sort_servers(servers, sort)
	store_servers(cache, servers)
	# display servers
	if format:
		# filter servers
		deletes = []
		if filter:
			regexp = compile(filter)
		for (i, server) in enumerate(servers):
			if not empty and server.is_empty():
				deletes.append(i)
				continue
			if not full and server.is_full():
				deletes.append(i)
				continue
			if filter:
				found = False
				for player in server.players:
					if regexp.search(decolorise(player.name)):
						found = True
						break
				if not found:
					deletes.append(i)
					continue
		deletes.reverse()
		for i in deletes:
			del servers[i]
		# print servers
		try:
			stdout.write(servers2text(servers, properties, players, format))
		except IOError, e:
			warn(str(e))

def list_servers(master, port, protocol, empty, full, cache,
		timeout, filter, format, properties, players, sort):
	"""List servers from master
	master		: the master address
	port		: the master port
	protocol	: the protocol of servers
	empty		: whether display or not empty servers
	full		: whether display or not full servers
	cache		: the cache on which to store servers
	timeout		: the time to wait for response from network
	filter		: the player name regexp to apply as filter on servers
	format		: the chosen output format
	properties	: whether display or not properties
	players		: whether display or not players
	sort		: the sort to use"""
	try:
		deferred = defer.Deferred()
		proto = Protocol(deferred)
		deferred.addCallback(digest_servers, cache, True,
				empty, full, filter, format, properties, players, sort)
		reactor.listenUDP(0, proto)
		proto.request_servers(Host(master, port), protocol, empty, full)
		reactor.callLater(timeout, proto.end)
		reactor.run()
	except Exception, e:
		raise QError(str(e))

def refresh_servers(cache, timeout, empty, full,
		filter, format, properties, players, sort):
	"""Refresh servers
	cache		: the cache on which to store servers
	timeout		: the time to wait for response from network
	empty		: whether display or not empty servers
	full		: whether display or not full servers
	filter		: the player name regexp to apply as filter on servers
	format		: the chosen output format
	properties	: whether display or not properties
	players		: whether display or not players
	sort		: the sort to use"""
	servers = load_servers(cache)
	if len(servers) != 0:
		try:
			deferred = defer.Deferred()
			proto = Protocol(deferred)
			deferred.addCallback(digest_servers, cache, False,
					empty, full, filter, format, properties, players, sort)
			reactor.listenUDP(0, proto)
			for server in servers:
				proto.request_server(server)
			reactor.callLater(timeout, proto.end)
			reactor.run()
		except Exception, e:
			raise QError(str(e))

def get_server(server, port, cache, timeout,
		format, properties, players):
	"""Get server
	server		: the server address
	port		: the server port
	cache		: the cache on which to store server
	timeout		: the time to wait for response from network
	format		: the chosen output format
	properties	: whether display or not properties
	players		: whether display or not players"""
	try:
		deferred = defer.Deferred()
		proto = Protocol(deferred)
		deferred.addCallback(digest_servers, cache, True,
				True, True, None, format, properties, players, None)
		reactor.listenUDP(0, proto)
		proto.request_server(Host(server, port))
		reactor.callLater(timeout, proto.end)
		reactor.run()
	except Exception, e:
		raise QError(str(e))

def add_favorite(cache, server, port):
	"""Add a server to favorites
	cache	: the cache on which to add favorite
	server	: the server address
	port	: the server port"""
	if path.exists(cache):
		servers = load_servers(cache)
	else:
		servers = []
	key = Host(server, port).get_key()
	for serv in servers:
		if serv.get_key() == key:
			raise QError("%s: server already in favorites" % str(key))
	servers.append(Server(server, port, {}, []))
	store_servers(cache, servers)

def delete_favorite(cache, server, port):
	"""Delete a server from favorites
	cache	: the cache from which to delete favorite
	server	: the server address
	port	: the server port"""
	servers = load_servers(cache)
	key = Host(server, port).get_key()
	found = False
	for (i, server) in enumerate(servers):
		if server.get_key() == key:
			del servers[i]
			found = True
			break
	if not found:
		raise QError("%s: server not found in favorites" % str(key))
	store_servers(cache, servers)

def load_configuration(file):
	"""Load configuration
	file	: the file from which to load configuration"""
	try:
		f = open(file)
	except IOError, e:
		raise QError(str(e))
	try:
		conf = f.readlines()
	except IOError, e:
		f.close()
		raise QError(str(e))
	f.close()
	for (n, line) in enumerate(conf):
		if IGNORE.match(line):
			continue
		m = PROPERTY.match(line)
		if m:
			key = m.group(1)
			if key in PROPERTIES:
				CONF[key] = m.group(2)
			else:
				raise QError("%s: %s: unknown option at line %d" \
						% (file, key, n + 1, ))
		else:
			raise QError("%s: bad option format at line %d" \
					% (file, n + 1, ))

def install():
	"""Install"""
	if not path.isdir(HOMEDIR):
		try:
			mkdir(HOMEDIR)
		except OSError, e:
			raise QError(str(e))

def check_address(address):
	"""Check address
	address	: the address to check"""
	m = ADDRESS.match(address)
	if m:
		address = m.group(1)
		port = m.group(3)
	else:
		port = None
	try:
		return (socket.gethostbyname(address), port)
	except socket.error, e:
		error("%s: %s" % (address, str(e)), 1)

def check_cache(cache, key):
	"""Check cache
	cache	: the cache to check
	key		: the cache key in configuration"""
	if not cache:
		if CONF.has_key(key):
			cache = CONF[key]
		else:
			error("cache required", 1)
	return path.expanduser(cache)

def check_master(master):
	"""Check master
	master	: the master to check"""
	if not master:
		if CONF.has_key("master"):
			master =  CONF["master"]
		else:
			error("master required", 1)
	return master

def check_filter(filter):
	"""Check filter
	filter	: the filter to check"""
	if not filter:
		if CONF.has_key("filter"):
			filter = CONF["filter"]
	return filter

def check_format(format, T, c, R, X, q):
	"""Check format
	format	: the default format
	T		: the text option
	c		: the color option
	R		: the raw option
	X		: the XML option
	q		: the quiet option"""
	if T + c + R + X + q > 1:
		error("only one format option allowed, try '%s --help'" % BASENAME, 1)
	if T:
		format = "text"
	elif c:
		format = "color"
	elif R:
		format = "raw"
	elif X:
		format = "xml"
	elif q:
		format = None
	elif CONF.has_key("format"):
		format = CONF["format"]
		if not FORMAT.match(format):
			error("%s: bad format" % format, 1)
	return format

def check_game(game):
	"""Check game
	game	: the game to check"""
	if game:
		conf = "%s.conf" % path.join(HOMEDIR, game)
		if path.exists(conf):
			try:
				# use 'game.conf' file in configuration directory
				load_configuration(conf)
			except QError, e:
				error(e.message, 1)
		else:
			try:
				# use 'game' file in current directory
				load_configuration(game)
			except QError, e:
				error(e.message, 1)
	else:
		game = "%s.conf" % path.join(HOMEDIR, BASENAME)
		if path.exists(game):
			try:
				# use 'script-name' file in configuration directory
				load_configuration(game)
			except QError, e:
				error(e.message, 1)
	return game

def check_port(port, key):
	"""Check port
	port	: the port to check
	key		: the port key in configuration"""
	if not port:
		if CONF.has_key(key):
			port = CONF[key]
		else:
			error("port required", 1)
	if NON_NEGATIVE_INTEGER.match(port):
		return int(port)
	else:
		error("%s: bad port number" % port, 1)

def check_protocol(protocol):
	"""Check protocol
	protocol	: the protocol to check"""
	if not protocol:
		if CONF.has_key("protocol"):
			protocol = CONF["protocol"]
		else:
			error("protocol required", 1)
	if POSITIVE_INTEGER.match(protocol):
		return int(protocol)
	else:
		error("%s: bad protocol number" % protocol, 1)

def check_timeout(timeout):
	"""Check timeout
	timeout	: the timeout to check"""
	if not timeout:
		if CONF.has_key("timeout"):
			timeout = CONF["timeout"]
		else:
			error("timeout required", 1)
	if NON_NEGATIVE_FLOAT.match(timeout):
		return float(timeout)
	else:
		error("%s: bad timeout number" % timeout, 1)

def check_sort(sort):
	"""Check sort
	sort	: the sort to check"""
	if not sort:
		if CONF.has_key("sort"):
			sort = CONF["sort"]
	if sort and not SORT.match(sort):
		error("%s: bad sort" % sort, 1)
	else:
		return sort

def main():
	"""Run it"""
	# default values
	A = None
	C = None
	c = False
	D = None
	E = False
	F = False
	f = False
	g = None
	h = False
	i = None
	l = False
	m = None
	P = None
	p = None
	q = False
	R = False
	r = False
	S = None
	s = None
	T = False
	t = False
	v = False
	W = None
	X = False
	y = False
	format = "text"
	# parse options
	try:
		options, args = getopt(argv[1:], "A:C:cD:EFg:fhi:lm:P:p:qRrS:s:TtvW:Xy",
				["add-favorite=", "cache=", "color", "delete-favorite=",
					"empty", "full", "favorites", "help", "filter=", "game=",
					"list", "master=", "protocol=", "port=", "quiet",
					"raw", "refresh", "sort=", "server=", "text",
					"properties", "version", "timeout=", "xml", "players"])
	except GetoptError:
		error("bad option, try '%s --help'" % BASENAME, 1)
	if args:
		error("bad argument, try '%s --help'" % BASENAME, 1)
	for option, arg in options:
		if option in ("-A", "--add-favorite"): A = arg
		elif option in ("-C", "--cache"): C = arg
		elif option in ("-c", "--color"): c = True
		elif option in ("-D", "--delete-favorite"): D = arg
		elif option in ("-E", "--empty"): E = True
		elif option in ("-F", "--full"): F = True
		elif option in ("-f", "--favorites"): f = True
		elif option in ("-g", "--game"): g = arg
		elif option in ("-h", "--help"): h = True
		elif option in ("-i", "--filter"): i = arg
		elif option in ("-l", "--list"): l = True
		elif option in ("-m", "--master"): m = arg
		elif option in ("-P", "--protocol"): P = arg
		elif option in ("-p", "--port"): p = arg
		elif option in ("-q", "--quiet"): q = True
		elif option in ("-R", "--raw"): R = True
		elif option in ("-r", "--refresh"): r = True
		elif option in ("-S", "--sort"): S = arg
		elif option in ("-s", "--server"): s = arg
		elif option in ("-T", "--text"): T = True
		elif option in ("-t", "--properties"): t = True
		elif option in ("-v", "--version"): v = True
		elif option in ("-W", "--timeout"): W = arg
		elif option in ("-X", "--xml"): X = True
		elif option in ("-y", "--players"): y = True
	# check arguments
	check = (A is not None) + (D is not None) + f \
			+ h + l + r + (s is not None) + v
	if check == 0:
		error("main option required, try '%s --help'" % BASENAME, 1)
	if check > 1:
		error("only one main option allowed, try '%s --help'" % BASENAME, 1)
	# display help
	if h:
		help()
		return
	if v:
		version()
		return
	# install
	try:
		install()
	except QError, e:
		error(e.message, 1)
	g = check_game(g)
	# add favorite
	if A:
		if c + E + F + (i is not None) + (m is not None) + (P is not None) \
				+ q + R + T + t + (W is not None) + X + y + (S is not None):
			error("unexpected option, try '%s --help'" % BASENAME, 1)
		(A, port) = check_address(A)
		if port:
			if p:
				error("port already specifed", 1)
			else:
				p = port
		C = check_cache(C, "cache_favorites")
		p = check_port(p, "port_server")
		try:
			add_favorite(C, A, p)
		except QError, e:
			error(e.message, 1)
		info("server %s:%d added to favorites" % (A, p))
		return
	# delete favorite
	if D:
		if c + E + F + (i is not None) + (m is not None) + (P is not None) \
				+ q + R + T + t + (W is not None) + X + y + (S is not None):
			error("unexpected option, try '%s --help'" % BASENAME, 1)
		(D, port) = check_address(D)
		if port:
			if p:
				error("port already specifed", 1)
			else:
				p = port
		C = check_cache(C, "cache_favorites")
		p = check_port(p, "port_server")
		try:
			delete_favorite(C, D, p)
		except QError, e:
			error(e.message, 1)
		info("server %s:%d deleted from favorites" % (D, p))
		return
	# list favorites
	if f:
		if (m is not None) + (P is not None) + (p is not None):
			error("unexpected option, try '%s --help'" % BASENAME, 1)
		C = check_cache(C, "cache_favorites")
		i = check_filter(i)
		W = check_timeout(W)
		S = check_sort(S)
		format = check_format(format, T, c, R, X, q)
		try:
			refresh_servers(C, W, E, F, i, format, t, y, S)
		except QError, e:
			error(e.message, 1)
		return
	# list servers
	if l:
		C = check_cache(C, "cache_servers")
		i = check_filter(i)
		P = check_protocol(P)
		W = check_timeout(W)
		S = check_sort(S)
		(m, port) = check_address(check_master(m))
		if port:
			if p:
				error("port already specifed", 1)
			else:
				p = port
		p = check_port(p, "port_master")
		format = check_format(format, T, c, R, X, q)
		try:
			list_servers(m, p, P, E, F, C, W, i, format, t, y, S)
		except QError, e:
			error(e.message, 1)
		return
	# refresh servers
	if r:
		if (m is not None) + (P is not None) + (p is not None):
			error("unexpected option, try '%s --help'" % BASENAME, 1)
		C = check_cache(C, "cache_servers")
		i = check_filter(i)
		W = check_timeout(W)
		S = check_sort(S)
		format = check_format(format, T, c, R, X, q)
		try:
			refresh_servers(C, W, E, F, i, format, t, y, S)
		except QError, e:
			error(e.message, 1)
		return
	# get server
	if s:
		if E + F + (i is not None) + (m is not None) + \
				(P is not None) + (S is not None):
			error("unexpected option, try '%s --help'" % BASENAME, 1)
		(s, port) = check_address(s)
		if port:
			if p:
				error("port already specifed", 1)
			else:
				p = port
		C = check_cache(C, "cache_server")
		W = check_timeout(W)
		p = check_port(p, "port_server")
		format = check_format(format, T, c, R, X, q)
		try:
			get_server(s, p, C, W, format, t, y)
		except QError, e:
			error(e.message, 1)
		return

if __name__ == "__main__":
	try:
		import psyco
		psyco.full()
	except ImportError:
		pass
	main()

